import numpy as np
import pandas as pd
import optuna
import os
import catboost as cat
from catboost import Pool
from .score_cal import RScore
import fasttreeshap
import matplotlib.pyplot as plt
from sklearn.model_selection import train_test_split

seed=42
def Train_CATR(features:any=None,
               target:any=None,
               iterations:int=50,
               scoring:str=None,
               validation_size:float=0.2,
               test_size:float=0.1,
               validation_set:tuple=None,
               test_set:tuple=None,
               task:str="MAE",
               objectives:any="valid_score", #{0: "valid_score", 1: "train_test_drop"}
               base_score:float=None,
               show_shap:bool=False,
               saved_dir:any=None,
               refit=False,
               ):
    """Construct a function to train a booster.

        Parameters
        ----------
        features : any, {dataframe, array} of predictors
            features use to build model.
        target : any, {array, Series} of target
            continuous target to be predicted.
        iterations : int,
            number of optimization rounds.
        scoring : str,
            scoring method for optimization.
            Available scoring method is sklearn regression metrics: R2 (r2), MAE, MSE, RMSE, MAPE, Poisson, Tweedie, MeAE,
                ExVS, MSLE, ME, Gama, D2T, D2Pi, D2A.
        validation_size: float,
            size of validation set use to evaluate model.
        task : str,
            internal method for catboost to optimize gradient. Available tasks of CatBoost regression are:
            MAE, MAPE, Poisson, Quantile, RMSE, Tweedie, R2, MSLE, MedianAbsoluteError, Huber.
        objectives : any, {int or str}
            the way to perform optimization.
            "valid_score" or 0: mean to optimize valid score
            "train_test_drop" or 1: mean to optimize different between train and test score.
        base_score: float
            control underfitting when apply objective is train_test_drop / 1
        start : any,
            default=1
            is internal param for starting optimize.
        show_shap : bool,
            default=False, show shap chart after tuning if True other while save charts as pictures.

        Returns
        -------
        estimator : object
            The trained model.
    """
    if validation_set and test_set:
        train_x, train_target = features, target
        valid_x, valid_target = validation_set
        test_x, test_target = test_set
        
    elif validation_set and test_size:
        valid_x, valid_target = validation_set
        train_x, test_x, train_target, test_target = train_test_split(features, 
                                                                      target, 
                                                                      test_size=test_size, 
                                                                      random_state=seed)
        
    elif test_set and validation_size:
        train_x, valid_x, train_target, valid_target = train_test_split(features, 
                                                                        target, 
                                                                        test_size=validation_size, 
                                                                        random_state=seed)
        test_x, test_target = test_set
        
    else:
        train_x, valid_x, train_target, valid_target = train_test_split(features, 
                                                                        target, 
                                                                        test_size=validation_size, 
                                                                        random_state=seed)
        
        train_x, test_x, train_target, test_target = train_test_split(features, 
                                                                      target, 
                                                                      test_size=test_size, 
                                                                      random_state=seed)
        
    # if validation_set:
    #     train_x, train_target = features, target
    #     valid_x, valid_target = validation_set
    # else:
    #     train_x, valid_x, train_target, valid_target = train_test_split(features, target, test_size=validation_size, random_state=seed)
    dtrain = Pool(data=train_x, label=train_target, feature_names=features.columns.to_list())
    dvalid = Pool(data=valid_x, label=valid_target, feature_names=features.columns.to_list())
    dtest  = Pool(data=test_x, label=test_target, feature_names=features.columns.to_list())

    def objective(trial):
        param = {
            "objective": task,
            "eval_metric": task,
            "bootstrap_type": trial.suggest_categorical("bootstrap_type", 
                                                        ["MVS", 
                                                         "Bernoulli", 
                                                         "Bayesian"]),
            "l2_leaf_reg": trial.suggest_loguniform("l2_leaf_reg", 1e-2, 5),
            "learning_rate": trial.suggest_loguniform("learning_rate", 0.001, 0.1),
            "mvs_reg": trial.suggest_float("mvs_reg", 0, 10),
            "min_data_in_leaf": trial.suggest_int("min_data_in_leaf", 10, 1000),
            "grow_policy": trial.suggest_categorical("grow_policy", 
                                                     ["SymmetricTree", 
                                                      "Lossguide", 
                                                      "Depthwise"]),
            "nan_mode": trial.suggest_categorical("nan_mode", ["Min", "Max"]),
        }

        if param["bootstrap_type"] != "Bayesian":
            param["subsample"] = trial.suggest_loguniform("subsample", 0.4, .9)
        if param["grow_policy"] == "Lossguide":
            param["max_leaves"] = trial.suggest_int("max_leaves", 11, 63)
        if param["grow_policy"] == "SymmetricTree":
            param["boosting_type"] = trial.suggest_categorical("boosting_type", 
                                                               ["Ordered", 
                                                                "Plain"])

        model = cat.train(
            pool=dtrain, 
            params=param,
            iterations=150,
            # eval_set=dvalid,
            # early_stopping_rounds=5,
            verbose=0,
            )
        #model.fit(train_x, train_labels)#, valid_sets=dvalid, num_boost_round=100)
        preds = model.predict(dvalid)
        pred_trains = model.predict(dtrain)
        valid_score = RScore(valid_target, preds, scoring=scoring)
        train_score = RScore(train_target, pred_trains, scoring=scoring)
        diff = train_score-valid_score
        #scores = ["R2", "ExV", "Poisson", "D2T", "D2Pi", "D2A"]
        if objectives==0 or objectives=='valid_score':
            print(f"train-test differs: {diff}")
            return valid_score
        else:
            if valid_score >= base_score:
                print(valid_score)
                return np.abs(diff)
            else:
                return np.exp(np.abs(diff)) * 10

    guides = ["R2", "ExV", "Poisson", "D2T", "D2Pi", "D2A"]

    if scoring in guides:
        if objectives==0 or objectives=='valid_score':
            direction="maximize"
        else: direction='minimize'
    else:
        direction="minimize"

    study = optuna.create_study(direction=direction, study_name="find best model")
    study.optimize(objective, n_trials=iterations)

    print("Number of finished trials: {}".format(len(study.trials)))

    print("Best trial:")
    trial = study.best_trial

    print("  Value: {}".format(trial.value))

    print("  Params: ")
    for key, value in trial.params.items():
        print("    {}: {}".format(key, value))

    best_params = {
            "objective": task,
            "eval_metric": task,
        }
    best_params.update(study.best_params)
    
    if refit:
        x_ = pd.concat([train_x, valid_x])
        target_ = pd.concat([train_target, valid_target])
        model = cat.train(
            pool=Pool(x_, label=target_), 
            params=best_params, 
            eval_set=dvalid, 
            verbose=0, 
            early_stopping_rounds=5,
            )
            
    else:
        model = cat.train(
            pool=dtrain, 
            params=best_params, 
            eval_set=dvalid, 
            verbose=0, 
            early_stopping_rounds=5,
            )

    preds = model.predict(dvalid)
    pred_trains = model.predict(dtrain)
    pred_tests = model.predict(dtest)
    
    valid_score = RScore(valid_target, 
                         preds, 
                         scoring=scoring)
    
    train_score = RScore(train_target, 
                         pred_trains, 
                         scoring=scoring)
    
    test_score  = RScore(test_target, 
                         pred_tests, 
                         scoring=scoring)
    
    print(f"Train score: {train_score}")
    if not refit:
        print(f"Valid score: {valid_score}")
    print(f"Test score: {test_score}")
        
    shap_values = model.get_feature_importance(Pool(features, label=target), 
                                               type='ShapValues')
    #expected_value = shap_values[0, -1]
    shap_values = shap_values[:, :-1]
    if saved_dir is None:
        list_dirs = os.scandir(os.getcwd())
        gate_ = [i for i in list_dirs if i.name=="catboost_shap"]
        if len(gate_) == 1:
            saved_dir = os.path.join(os.getcwd(), "catboost_shap")
        else:
            os.mkdir(os.path.join(os.getcwd(), "catboost_shap"))
            saved_dir = os.path.join(os.getcwd(), "catboost_shap")

    if show_shap:
        fasttreeshap.summary_plot(shap_values, features)
        for feature in features.columns:
            fasttreeshap.dependence_plot(feature, shap_values, features)
    else:
        fasttreeshap.summary_plot(shap_values, features, show=False)
        f = plt.gcf()
        plt.title(f"Feature importances of Catboost model")
        plt.savefig(f"{saved_dir}/Catboost_summary_plot.png", bbox_inches='tight')
        plt.close()
        # visualize the first prediction's explanation
        for feature in features.columns:
            fasttreeshap.dependence_plot(feature, 
                                         shap_values, 
                                         features, 
                                         show=False)
            f = plt.gcf()
            plt.title(f"Catboost_{feature}_features dependency plot")
            plt.savefig(f"{saved_dir}/Catboost_{feature}_features dependency plot.png",
                        bbox_inches='tight')
            plt.close()
    return model



import pandas as pd
import plotly.express as px
import seaborn as sns
import altair as alt

#Histogram Line-------------------------------------------------------------------------------------------------------
def linehist(data, curve):
    if curve in ["LLD", "LLS"]:
        fig = sns.displot(data, x=curve, hue="WELL", kind="kde", height=5,aspect=1.2, log_scale=True)
        fig.set(ylabel="Values")
    else:
        fig = sns.displot(data, x=curve, hue="WELL", kind="kde", height=5,aspect=1.2)
        fig.set(ylabel="Values")
        
#CrossPlot------------------------------------------------------------------------------------------------------------
def cross_plot(data, x_axis, y_axis, log_x:bool=False, log_y:bool=False, hue:str=None):     
    fig = sns.jointplot(data=data, x=x_axis, y=y_axis, hue=hue)
    if log_x:
        fig.ax_joint.set_xscale('log')
        fig.ax_marg_x.set_xscale('log')
    if log_y:
        fig.ax_joint.set_yscale('log')
        fig.ax_marg_y.set_yscale('log')
        
#PairPlot-------------------------------------------------------------------------------------------------------------
def pair_plot(data, rows, cols,color_):
    return alt.Chart(data).mark_circle().encode(
            alt.X(alt.repeat("column"), type='quantitative', scale=alt.Scale(zero=False)),
            alt.Y(alt.repeat("row"), type='quantitative', scale=alt.Scale(zero=False)),
            color=color_
            ).properties(
                width=100,
                height=100
            ).repeat(
                row = rows,
                column = cols
            ).configure_axis(
                grid=False
            )
#Heatmap--------------------------------------------------------------------------------------------------------------
def heatmap(df):            
    fig = sns.heatmap(df, annot=True)
    
#3D-------------------------------------------------------------------------------------------------------------------
def scatter_3d(data:pd.DataFrame = None, 
              x_value: str = None, 
              y_value: str = None, 
              z_value: str = None, 
              color: str = None, 
              opacity : float = 1,
              symnbol_size : str = None, 
              size_max : int = 18,
              symbol : str = None, 
              log_x : bool = False, 
              log_y : bool = False, 
              log_z : bool = False,
              size: list = [1000, 700],
              ):
    
    fig = px.scatter_3d(data, 
                        x=x_value, 
                        y=y_value, 
                        z=z_value,
                        color=color, 
                        size=symnbol_size, 
                        size_max=size_max,
                        symbol=symbol, 
                        opacity=opacity,
                        log_x=log_x,
                        log_y=log_y,
                        log_z=log_z,
                        width=size[0], height=size[1],
                        color_continuous_scale="blugrn",
                        )
    fig.update_layout(margin=dict(l=0, r=0, b=0, t=0), #tight layout
                    #   paper_bgcolor="LightSteelBlue"
                    template="none")
    fig.show()


import numpy as np
import pandas as pd
import optuna
import xgboost as xgb
from .score_cal import RScore
from sklearn.model_selection import train_test_split
import os
seed = 42


def Train_XGBR(features: any = None,
               target: any = None,
               iterations: int = 50,
               scoring: str = None,
               validation_size: float = 0.2,
               test_size: float = 0.1,
               validation_set: tuple = None,
               test_set: tuple = None,
               task: str = "reg:squarederror",
               base_score:float=None,
               # {0: "valid_score", 1: "train_valid_drop"}
               objectives: any = "valid_score",
               start: any = 1,  # 1 or True
               show_shap: bool = False,
               saved_dir:any=None,
               refit: bool = False,
               ):
    """Construct a function to train a booster.

        Parameters
        ----------
        features : any, {dataframe, array} of predictors
            features use to build model.
        target : any, {array, Series} of target
            continuous target to be predicted.
        iterations : int,
            number of optimization rounds.
        scoring : str,
            scoring method for optimization.
            Available scoring method is sklearn regression metrics: R2 (r2), MAE, MSE, RMSE, MAPE, Poisson, Tweedie, MeAE,
                ExVS, MSLE, ME, Gama, D2T, D2Pi, D2A.
        validation_size: float,
            size of validation set use to evaluate model.
        test_size: float,
            size of test set.
        validation_set: tuple,
            tuple of (X_valid, y_valid)
        test_set: tuple,
            tuple of (X_test, y_test)
        task : str,
            internal method for xgboost to optimize gradient. Available tasks of xgboost regression are:
            reg:squarederror, reg:squaredlogerror, reg:pseudohubererror, count:poisson, survival:cox, survival:aft,
            reg:gamma, reg:tweedie, rank:ndcg, rank:pairwise, rank:map.
        objectives : any, {int or str}
            the way to perform optimization.
            "valid_score" or 0: mean to optimize valid score
            "train_test_drop" or 1: mean to optimize different between train and valid score
        base_score: float
            control under fitting when apply objective is train_test_drop / 1
        start : any,
            default=1
            is internal param for starting optimize
        show_shap : bool,
            default=False, show shap chart after tuning if True other while save charts as pictures

        Returns
        -------
        estimator : object
            The trained model.
        """

    if validation_set and test_set:
        train_x, train_target = features, target
        valid_x, valid_target = validation_set
        test_x, test_target = test_set

    elif validation_set and test_size:
        valid_x, valid_target = validation_set
        train_x, test_x, train_target, test_target = train_test_split(features,
                                                                      target,
                                                                      test_size=test_size,
                                                                      random_state=seed)

    elif test_set and validation_size:
        train_x, valid_x, train_target, valid_target = train_test_split(features,
                                                                        target,
                                                                        test_size=validation_size,
                                                                        random_state=seed)
        test_x, test_target = test_set

    else:
        train_x, valid_x, train_target, valid_target = train_test_split(features,
                                                                        target,
                                                                        test_size=validation_size,
                                                                        random_state=seed)

        train_x, test_x, train_target, test_target = train_test_split(features,
                                                                      target,
                                                                      test_size=test_size,
                                                                      random_state=seed)

    # if validation_set:
        # train_x, train_target = features, target
        # valid_x, valid_target = validation_set
    # else:
        # train_x, valid_x, train_target, valid_target = train_test_split(
        # features, target, test_size=validation_size, random_state=seed)
    dtrain = xgb.DMatrix(train_x, label=train_target)
    dvalid = xgb.DMatrix(valid_x, label=valid_target)
    dtest = xgb.DMatrix(test_x, label=test_target)

    def objective(trial):
        param = {
            "objective": task,
            "booster": trial.suggest_categorical("booster", ["gbtree", "gblinear"]),
            "lambda": trial.suggest_loguniform("lambda", 1e-2, 5.0),
            "alpha": trial.suggest_loguniform("alpha", 1e-2, 5.0),
            "learning_rate": trial.suggest_loguniform("learning_rate", 1e-1, 0.10),
        }

        if param["booster"] == "gbtree":
            param["colsample_bytree"] = trial.suggest_loguniform(
                "colsample_bytree", 0.5, 0.9)
            param["colsample_bynode"] = trial.suggest_loguniform(
                "colsample_bynode", 0.5, 0.9)
            param["colsample_bylevel"] = trial.suggest_loguniform(
                "colsample_bylevel", 0.5, 0.9)
            param["gamma"] = trial.suggest_loguniform("gamma", 1e-2, 10.0)
            param["num_parallel_tree"] = trial.suggest_int(
                "num_parallel_tree", 1, 5)
            param["subsample"] = trial.suggest_loguniform(
                "subsample", 0.5, 0.9)
            param["tree_method"] = trial.suggest_categorical(
                "tree_method", ["exact", "hist"])
            param["min_child_weight"] = trial.suggest_int(
                "min_child_weight", 5, 10)
            if param["tree_method"] != "exact":
                param["max_leaves"] = trial.suggest_categorical(
                    "max_leaves", [21, 31, 41, 51, 61])
                param["max_bin"] = trial.suggest_categorical(
                    "max_bin", [50, 100, 150, 200, 256])
                param["grow_policy"] = trial.suggest_categorical(
                    "grow_policy", ["depthwise", "lossguide"])
        if param["objective"] == "reg:tweedie":
            param["tweedie_variance_power"] = trial.suggest_loguniform(
                "tweedie_variance_power", 1., 1.99)
        if param["objective"] == "reg:pseudohubererror":
            param["huber_slope"] = trial.suggest_loguniform(
                "huber_slope", 0.5, 2)
        if param["booster"] == "gblinear":
            param["updater"] = trial.suggest_categorical(
                "updater", ["shotgun", "coord_descent"])
            param["feature_selector"] = trial.suggest_categorical(
                "feature_selector", ["cyclic", "shuffle"])

        model = xgb.train(
            param,
            dtrain,
            # evals=[(dtrain, "train"), (dvalid, "valid")],
            num_boost_round=100,
            # early_stopping_rounds=5,
            verbose_eval=0
        )
        preds = model.predict(dvalid)
        pred_trains = model.predict(dtrain)
        valid_score = RScore(valid_target, preds, scoring=scoring)
        train_score = RScore(train_target, pred_trains, scoring=scoring)
        diff = train_score-valid_score
        if objectives == 0 or objectives == 'valid_score':
            print(f"train-test differs: {diff}")
            return valid_score
        else:
            if valid_score >= base_score:
                print(valid_score)
                return np.abs(diff)
            else:
                return np.exp(np.abs(diff)) * 10

    guides = ["R2", "ExV", "Poisson", "D2T", "D2Pi", "D2A"]

    if scoring in guides:
        if objectives == 0 or objectives == 'valid_score':
            direction = "maximize"
    else:
        direction = "minimize"

    if start == 1 or start == True:
        study = optuna.create_study(
            direction=direction, study_name="find best model")
        study.optimize(objective, n_trials=iterations)

        print("Number of finished trials: {}".format(len(study.trials)))

        print("Best trial:")
        trial = study.best_trial

        print("  Value: {}".format(trial.value))

        print("  Params: ")
        for key, value in trial.params.items():
            print("    {}: {}".format(key, value))

    best_params = {
        "objective": task,
        "eval_metric": "mae",
    }

    best_params.update(study.best_params)
    if refit:
        x_ = pd.concat([train_x, valid_x])
        target_ = pd.concat([train_target, valid_target])
        model = xgb.train(
            best_params,
            xgb.DMatrix(x_, label=target_),
            num_boost_round=500,
            early_stopping_rounds=10,
            evals=[(dtrain, 'train'), (dvalid, 'valid')],
            verbose_eval=False
        )
    else:
        model = xgb.train(
            best_params,
            dtrain,
            num_boost_round=500,
            early_stopping_rounds=10,
            evals=[(dtrain, 'train'), (dvalid, 'valid')],
            verbose_eval=False
        )

    preds = model.predict(dvalid)
    pred_trains = model.predict(dtrain)
    pred_tests = model.predict(dtest)
    valid_score = RScore(valid_target, preds, scoring=scoring)
    train_score = RScore(train_target, pred_trains, scoring=scoring)
    test_score = RScore(test_target, pred_tests, scoring=scoring)
    print(f"Train score: {train_score}")
    if not refit:
        print(f"Valid score: {valid_score}")
    print(f"Test score: {test_score}")

    if saved_dir is None:
        list_dirs = os.scandir(os.getcwd())
        gate_ = [i for i in list_dirs if i.name=="xgboost_shap"]
        if len(gate_) == 1:
            saved_dir = os.path.join(os.getcwd(), "xgboost_shap")
        else:
            os.mkdir(os.path.join(os.getcwd(), "xgboost_shap"))
            saved_dir = os.path.join(os.getcwd(), "xgboost_shap")

    if best_params["booster"] != "gblinear":
        if show_shap:
            import sys
            sys.path.append('../Shapley')
            from .ShapleyC import shapley_importances_C
            shapley_importances_C(model=model, X=features,
                                feature_names=features.columns,
                                shap_sample_size=1., show_plot=show_shap, saved_dir=saved_dir)

        else:
            import sys
            sys.path.append('../Shapley')
            from .ShapleyC import shapley_importances_C
            shapley_importances_C(
                model=model, X=features,
                feature_names=features.columns,
                shap_sample_size=1., show_plot=False, saved_dir=saved_dir)
    else:
        pass
    return model



import pandas as pd
import plotly.express as px

def missing_stats(data:pd.DataFrame = None,
                  title: str = None,
                  ):
    missing = data.isnull().sum()*100/data.isnull().sum().sum()
    missing = missing[missing >= 0].reset_index()
    missing.columns = ['Columns', 'Count missing (%)']
    
    fig = px.bar(missing, y='Count missing (%)', x='Columns', text_auto='.2s',
            title=title)
    fig.show()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Jul 30 21:16:45 2022

@author: nguyens
"""
import numpy as np
import pandas as pd
from sklearn.metrics import balanced_accuracy_score, fbeta_score, accuracy_score
from sklearn.metrics import recall_score, precision_score, roc_auc_score
from sklearn.metrics import r2_score, mean_absolute_error
from sklearn.metrics import mean_squared_error, explained_variance_score
from sklearn.metrics import mean_squared_log_error, mean_poisson_deviance, mean_absolute_percentage_error
from sklearn.metrics import median_absolute_error, max_error, mean_gamma_deviance, mean_tweedie_deviance
from sklearn.metrics import mean_pinball_loss, d2_tweedie_score, d2_pinball_score, d2_absolute_error_score

def Score(y_true:any=None, y_pred:any=None, scoring:str=None, favor_class:any=None):
    
    """
    This function use to calculate score of model based on input scoring method
     - Input:
             + y_true : a label series/array
             + y_pred: a predict series/array
             + scoring: Scoring method [balanced_accuracy, f1_weighted, 
                                        recall, precision, roc_auc]
             + favor_class: when using f1_weighted as scoring
     - Output: 
         + model score
    Example:
        #>>> from score_cal import Score
        #>>> score = Score(y_true, y_pred, scoring='f1_weighted', favor_class=1)
        
    """
    
    if scoring=="balanced_accuracy":
        score = balanced_accuracy_score(y_true, y_pred)

    elif scoring=="f1_weighted":
        if favor_class==1 or favor_class=="min false positive":
            score = fbeta_score(y_true, y_pred, beta=0.5, average='weighted')

        elif favor_class==0 or favor_class=="min false negative":
            score = fbeta_score(y_true, y_pred, beta=2, average='weighted')

        else:
            score = fbeta_score(y_true, y_pred, beta=1, average='weighted')

    elif scoring=="recall":
        score = recall_score(y_true, y_pred, average='weighted')

    elif scoring=="precision":
        score = precision_score(y_true, y_pred, average='weighted')

    elif scoring=="accuracy":
        score = accuracy_score(y_true, y_pred)

    else:
        score = roc_auc_score(y_true, y_pred, average='weighted')
        
    return score

# make scoring
def RScore(y_true, y_pred, scoring):
    score_dict = {
        "R2": r2_score,
        "MAE": mean_absolute_error,
        "MSE": mean_squared_error,
        "RMSE": root_mean_squared_error,
        "ExVS": explained_variance_score,
        "MSLE": mean_squared_log_error,
        "Poisson": mean_poisson_deviance,
        "MAPE": mean_absolute_percentage_error,
        "MeAE": median_absolute_error,
        "ME": max_error,
        "Gama": mean_gamma_deviance,
        "Tweedie": mean_tweedie_deviance,
        "Pinball": mean_pinball_loss,
        "D2T": d2_tweedie_score,
        "D2Pi": d2_pinball_score,
        "D2A": d2_absolute_error_score
    }

    score = score_dict[scoring]

    return score(y_true, y_pred)

def root_mean_squared_error(y_true, y_pred):
    return np.sqrt(mean_squared_error(y_true, y_pred))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Jul 30 21:14:18 2022

@author: nguyens
"""
import numpy as np
from sklearn.metrics import make_scorer, precision_score, roc_auc_score
from sklearn.metrics import balanced_accuracy_score, recall_score, fbeta_score

# make scoring
def My_Score(scoring:str=None, favor_class:any=1):
    score_dict = {
                    "balanced_accuracy": balanced_accuracy_score,
                    "recall": recall_score,
                    "f1_weighted": fbeta_score,
                    "precision": precision_score,
                    "roc_auc": roc_auc_score,
                    }

    if scoring == "roc_auc":
        my_score_ = make_scorer(score_func=score_dict[scoring], needs_proba=True, labels=[0, 1], average="weighted")

    elif scoring == "f1_weighted":
        if favor_class==1 or favor_class=="min false positive":
            my_score_ = make_scorer(score_func=score_dict[scoring], beta=0.5, average='weighted')

        elif favor_class==0 or favor_class=="min false negative":
            my_score_ = make_scorer(score_func=score_dict[scoring], beta=2, average='weighted')

        else:
            my_score_ = make_scorer(score_func=score_dict[scoring], beta=1, average='weighted')

    elif scoring in ["precision", "recall"]:
        my_score_ = make_scorer(score_func=score_dict[scoring])

    else:
        my_score_ = score_dict["balanced_accuracy"]

    return my_score_



import pandas as pd
import plotly.graph_objects as go
from plotly.subplots import make_subplots
import json
#Curve View--------------------------------------------------------------------------------------------------------------
#OPERATIONS--------------------------------------------------------------------------------------------------------------
#Make subplots layout
def view_curves(data:pd.DataFrame=None, curves:list=[], log:list=[],
         depth_col:str='DEPTH', new_color:dict={}, new_unit:dict={},
         ):
    #FUNCTIONS DESIGNING-------------------------------------------------------------------------------------------------
    #Function of Plotting single curve
    def add_trace(curv:str=None, color:str='#38a2e8', xaxis:str='x1', unit:str=None, id_subplt:int=1, log:list=None):
        fig.add_trace(go.Scattergl(y=data[depth_col], x=data[curv], name=curv, mode="lines", line=dict(color=color, width=1.), xaxis=xaxis),1 ,id_subplt)\
            .update_xaxes(title_text=f'{curv} ({unit})', row=1, col=id_subplt, color=color,
                          tickcolor=color, minor=dict(tickcolor=color), tickfont=dict(color=color))
        if curv in log:
            fig.update_xaxes(type='log', row=1, col=id_subplt)

    #Function to update X-Axis parameters
    def update_xaxes(type:str=None):
                fig.update_xaxes(
                                title_font=dict(size=16, family='Arial'),
                                title_standoff = 0, side='top', nticks=5, fixedrange=True,
                                ticks="inside", tickwidth=1, ticklen=8, ticklabelstep=1,
                                tickfont=dict(family="Arial", size=12),
                                minor_ticks="inside", minor=dict(ticklen=4, showgrid=True),
                                showline=True, linewidth=1, linecolor='black', mirror=True,
                                showgrid=True, gridwidth=1, gridcolor='#d9d9d9',
                                showspikes=True, spikecolor="#8d9199", spikesnap="hovered data", spikemode="across", spikethickness=1, spikedash='solid',
                                )
                if type is not None:
                        fig.update_xaxes(type=type)
                        
    #Function to update Y-Axis parameters
    def update_yaxes(type:str=None):
            fig.update_yaxes(
                            # title_font=dict(size=18, family='Arial', color='#393533'),
                            # matches='y'
                            title_standoff = 0,
                            autorange="reversed",
                            nticks=10,
                            ticks="outside", tickwidth=1, tickcolor='black', ticklen=8, ticklabelstep=1, 
                            tickfont=dict(family="Arial", color='black', size=12),
                            minor_ticks="outside", minor=dict(ticklen=4, tickcolor="black", showgrid=False),
                            showline=True, linewidth=1, linecolor='black', mirror=True,
                            showgrid=True, gridwidth=1, gridcolor='#d9d9d9',
                            rangemode='nonnegative', #(normal or tozero or nonnegative)
                            showspikes=True, spikecolor="#8d9199", spikesnap="cursor", spikemode="across", spikethickness=1, spikedash='solid',
                            )
            if type is not None:
                    fig.update_yaxes(type=type)

    #Function to update FIGURE LAYOUT parameters
    def update_layout(_size:list=[400,800]):
            fig.update_layout(
                            width=_size[0], height=_size[1],
                            showlegend=False,
                            plot_bgcolor="white",
                            margin=dict(t=10,l=10,b=10,r=10),
                            hovermode="y", #"y unified", "y", "x", "closest"
                            )

    #VARIABLES DEFINITION------------------------------------------------------------------------------------------------
    #Sort dataframe by DEPTH
    data = data.sort_values(by=depth_col)

    #Default color codes
    meta_info = '{"Unit":{"HCAL":"in","CALI":"in","CALS":"in","C1":"in","C2":"in","RHOB":"g\\/c3","RHO8":"g\\/c3","RHO8_filte":"g\\/c3","RHOZ":"g\\/c3","ZDEN":"g\\/c3","ZDNC":"g\\/c3","HDRA":"g\\/c3","DRHO":"g\\/c3","DSOZ":"g\\/c3","ECD_ARC":"g\\/c3","NPHI":"v\\/v","TNPH":"v\\/v","TNPH_CH":"v\\/v","HNPO":"v\\/v","HNPO_filte":"v\\/v","CNCF":"% (pu)","NPOR":"v\\/v","CH_NPHI":"v\\/v","CH_NPHIds":"v\\/v","CFTC":"Hz","CNTC":"Hz","DTC":"us\\/f","DTCO":"us\\/f","DT":"us\\/f","DT4P":"us\\/f","DT5":"us\\/f","DT_SPL":"us\\/f","DTS":"us\\/f","DTSM":"us\\/f","DT2":"us\\/f","DTSO":"us\\/f","DT4S":"us\\/f","GR":"GAPI","GR1":"GAPI","GR_ARC":"GAPI","CGR":"GAPI","EHGR":"GAPI","ECGR":"GAPI","SGR":"GAPI","GR_SPULSE_BHC":"GAPI","GR_IMP":"GAPI","GRM1":"GAPI","MWD_GR_BHC":"GAPI","GRAM":"GAPI","GR_SPL":"GAPI","GR_CDR":"GAPI","ARC_GR_RT":"GAPI","GR_LWD":"GAPI","POTA":"%","THOR":"ppm","URAN":"ppm","TPRA":"v\\/v","TURA":"v\\/v","UPRA":"v\\/v","LLD":"Ohmm","LLS":"Ohmm","DIL":"Ohmm","ILD":"Ohmm","ILM":"Ohmm","RILD":"Ohmm","RILM":"Ohmm","HRI":"Ohmm","RIDPH":"Ohmm","RIMPH":"Ohmm","RSFL":"Ohmm","HDRS":"Ohmm","HMRS":"Ohmm","LL3":"Ohmm","RIPD":"Ohmm","RIPM":"Ohmm","AT10":"Ohmm","AT20":"Ohmm","AT30":"Ohmm","AT60":"Ohmm","AT90":"Ohmm","AO10":"Ohmm","AO20":"Ohmm","AO30":"Ohmm","AO60":"Ohmm","AO90":"Ohmm","AF10":"Ohmm","AF20":"Ohmm","AF30":"Ohmm","AF60":"Ohmm","AF90":"Ohmm","AHO10":"Ohmm","AHO20":"Ohmm","AHO30":"Ohmm","AHO60":"Ohmm","AHO90":"Ohmm","M2RX":"Ohmm","M2R9":"Ohmm","M2R6":"Ohmm","M2R3":"Ohmm","M2R1":"Ohmm","P16H_UNC":"Ohmm","P22H_UNC":"Ohmm","P28H_UNC":"Ohmm","P34H_UNC":"Ohmm","P40H_UNC":"Ohmm","A28H_UNC":"Ohmm","A34H_UNC":"Ohmm","A40H_UNC":"Ohmm","P16L_UNC":"Ohmm","P22L_UNC":"Ohmm","P28L_UNC":"Ohmm","P34L_UNC":"Ohmm","P40L_UNC":"Ohmm","P22H_UNC_RT":"Ohmm","P40H_UNC_RT":"Ohmm","P40H_UNC_R":"Ohmm","P22H_UNC_R":"Ohmm","HLLD":"Ohmm","HLLS":"Ohmm","RES_DEEP":"Ohmm","RES_SHAL":"Ohmm","Rdeep":"Ohmm","Rshallow":"Ohmm","P33H_UNC":"Ohmm","P33L_UNC":"Ohmm","A33H_UNC":"Ohmm","A33L_UNC":"Ohmm","RACHM":"Ohmm","RACLM":"Ohmm","RPCHM":"Ohmm","RPCLM":"Ohmm","RPCSHM":"Ohmm","RPCSLM":"Ohmm","RACSHM":"Ohmm","RACSLM":"Ohmm","RAC1HM":"Ohmm","RAC1LM":"Ohmm","RPC1HM":"Ohmm","RPC1LM":"Ohmm","RAC2HM":"Ohmm","RAC2LM":"Ohmm","RPC2HM":"Ohmm","RPC2LM":"Ohmm","RAC3HM":"Ohmm","RAC3LM":"Ohmm","RPC3HM":"Ohmm","RPC3LM":"Ohmm","RAC4HM":"Ohmm","RAC4LM":"Ohmm","RPC4HM":"Ohmm","RPC4LM":"Ohmm","MSFL":"Ohmm","RXO8":"Ohmm","ATR":"Ohmm","PSR":"Ohmm","RT":"Ohmm","TAB_CDR_RES":"hr","TAB_ARC_RES":"s","TAB_RES":"s","PEF":"_","PEF8":"_","PERF":"_","PEFZ":"_","ROP5_RM":"m\\/hr","ROP":"m\\/hr","ROP5":"m\\/hr","ROPS":"m\\/hr","RXOZ":"Ohmm","RSOZ":"Ohmm","ITT":"s","TENS":"lbf","TTEN":"lbf","HTEN":"lbf","CDF":"lbf","ATMP":"degC","TEMP_MCR":"_","TVDE":"m","SP":"mV","VPVS":"v\\/v","LSN":"cps","SSN":"cps","APRS_ARC":"psi","ARTM":"_","AGTK":"_","ARTK":"_","Temp":"degC","TEMP_LWD":"degC","WTBH":"degC","WTBH1":"degC","TEMP_DNI_RT":"degC","HSTEMP":"degC","TCDM":"degC","RPTHM":"mins","CHT":"lbf","DPHI":"v\\/v","SPHI":"v\\/v","PHIE":"v\\/v","PR":"_","ICV":"m3","IHV":"m3","GTEM":"degC","DHTEMP":"degC","TTEM_CDR":"degC","SVEL":"m\\/s","TTSL":"us","TT":"us","CBSL":"mV","CBL":"mV","WF2":"_","WF1":"_","CBLF":"mV","CCL":"_","BS":"_","TGAS":"ppm","Oilshow":"_"},"Color":{"HCAL":"#bf2273","CALI":"#bf2273","CALS":"#bf2273","C1":"#bf2273","C2":"#bf2273","RHOB":"#f20a0a","RHO8":"#f20a0a","RHO8_filte":"#f20a0a","RHOZ":"#f20a0a","ZDEN":"#f20a0a","ZDNC":"#f20a0a","HDRA":"#f20a0a","DRHO":"#f20a0a","DSOZ":"#f20a0a","ECD_ARC":"#f20a0a","NPHI":"#0a44f2","TNPH":"#0a44f2","TNPH_CH":"#0a44f2","HNPO":"#0a44f2","HNPO_filte":"#0a44f2","CNCF":"#0a44f2","NPOR":"#0a44f2","CH_NPHI":"#0a44f2","CH_NPHIds":"#0a44f2","CFTC":"#0a44f2","CNTC":"#0a44f2","DTC":"#ea0af2","DTCO":"#ea0af3","DT":"#ea0af4","DT4P":"#ea0af5","DT5":"#ea0af6","DT_SPL":"#ea0af7","DTS":"#630af2","DTSM":"#630af3","DT2":"#630af4","DTSO":"#630af6","DT4S":"#630af7","GR":"#40f20a","GR1":"#40f20a","GR_ARC":"#40f20a","CGR":"#40f20a","EHGR":"#40f20a","ECGR":"#40f20a","SGR":"#40f20a","GR_SPULSE_BHC":"#40f20a","GR_IMP":"#40f20a","GRM1":"#40f20a","MWD_GR_BHC":"#40f20a","GRAM":"#40f20a","GR_SPL":"#40f20a","GR_CDR":"#40f20a","ARC_GR_RT":"#40f20a","GR_LWD":"#40f20a","POTA":"#0a0a0a","THOR":"#0a0a0a","URAN":"#0a0a0a","TPRA":"#0a0a0a","TURA":"#0a0a0a","UPRA":"#0a0a0a","LLD":"#f20a0a","LLS":"#0a44f2","DIL":"#0a44f2","ILD":"#f20a0a","ILM":"#0a44f2","RILD":"#f20a0a","RILM":"#0a44f2","HRI":"#f20a0a","RIDPH":"#0a44f2","RIMPH":"#eb0edc","RSFL":"#f20a0a","HDRS":"#0a44f2","HMRS":"#eb0edc","LL3":"#f20a0a","RIPD":"#0a44f2","RIPM":"#eb0edc","AT10":"#f20a0a","AT20":"#f20a0a","AT30":"#f20a0a","AT60":"#0a44f2","AT90":"#0a44f2","AO10":"#f20a0a","AO20":"#f20a0a","AO30":"#f20a0a","AO60":"#0a44f2","AO90":"#0a44f2","AF10":"#f20a0a","AF20":"#f20a0a","AF30":"#f20a0a","AF60":"#0a44f2","AF90":"#0a44f2","AHO10":"#f20a0a","AHO20":"#f20a0a","AHO30":"#f20a0a","AHO60":"#0a44f2","AHO90":"#0a44f2","M2RX":"#0a44f2","M2R9":"#0a44f2","M2R6":"#0a44f2","M2R3":"#f20a0a","M2R1":"#f20a0a","P16H_UNC":"#f20a0a","P22H_UNC":"#f20a0a","P28H_UNC":"#f20a0a","P34H_UNC":"#0a44f2","P40H_UNC":"#0a44f2","A28H_UNC":"#f20a0a","A34H_UNC":"#0a44f2","A40H_UNC":"#0a44f2","P16L_UNC":"#f20a0a","P22L_UNC":"#f20a0a","P28L_UNC":"#f20a0a","P34L_UNC":"#0a44f2","P40L_UNC":"#0a44f2","P22H_UNC_RT":"#f20a0a","P40H_UNC_RT":"#0a44f2","P40H_UNC_R":"#0a44f2","P22H_UNC_R":"#f20a0a","HLLD":"#f20a0a","HLLS":"#0a44f2","RES_DEEP":"#f20a0a","RES_SHAL":"#0a44f2","Rdeep":"#f20a0a","Rshallow":"#0a44f2","P33H_UNC":"#0a44f2","P33L_UNC":"#f20a0a","A33H_UNC":"#0a44f2","A33L_UNC":"#f20a0a","RACHM":"#0a44f2","RACLM":"#f20a0a","RPCHM":"#0a44f2","RPCLM":"#f20a0a","RPCSHM":"#0a44f2","RPCSLM":"#f20a0a","RACSHM":"#0a44f2","RACSLM":"#f20a0a","RAC1HM":"#0a44f2","RAC1LM":"#f20a0a","RPC1HM":"#0a44f2","RPC1LM":"#f20a0a","RAC2HM":"#0a44f2","RAC2LM":"#f20a0a","RPC2HM":"#0a44f2","RPC2LM":"#f20a0a","RAC3HM":"#0a44f2","RAC3LM":"#f20a0a","RPC3HM":"#0a44f2","RPC3LM":"#f20a0a","RAC4HM":"#0a44f2","RAC4LM":"#f20a0a","RPC4HM":"#0a44f2","RPC4LM":"#f20a0a","MSFL":"#0a44f2","RXO8":"#f20a0a","ATR":"#0a44f2","PSR":"#f20a0a","RT":"#f20a0a","TAB_CDR_RES":"#f20a0a","TAB_ARC_RES":"#f20a0a","TAB_RES":"#f20a0a","PEF":"#f70ad0","PEF8":"#f70ad1","PERF":"#f70ad2","PEFZ":"#f70ad3","ROP5_RM":"#f20a0a","ROP":"#f20a0a","ROP5":"#f20a0a","ROPS":"#f20a0a","RXOZ":"#0e33eb","RSOZ":"#0e33eb","ITT":"#2291f2","TENS":"#11f2f2","TTEN":"#11f2f3","HTEN":"#11f2f4","CDF":"#11f2f5","ATMP":"#fa0202","TEMP_MCR":"#fa0203","TVDE":"#0f1d29","SP":"#fa4402","VPVS":"#e102fa","LSN":"#ed0510","SSN":"#0533ed","APRS_ARC":"#ed0510","ARTM":"#f20a0a","AGTK":"#40f20a","ARTK":"#f20a0a","Temp":"#fa0202","TEMP_LWD":"#fa0203","WTBH":"#fa0204","WTBH1":"#fa0205","TEMP_DNI_RT":"#fa0206","HSTEMP":"#fa0207","TCDM":"#fa0208","RPTHM":"#454141","CHT":"#34fa02","DPHI":"#fa4402","SPHI":"#fa02e9","PHIE":"#fa0249","PR":"#fa02f6","ICV":"#9c959b","IHV":"#9c959b","GTEM":"#fa0202","DHTEMP":"#fa0203","TTEM_CDR":"#fa0205","SVEL":"#dd02fa","TTSL":"#dd02fa","TT":"#dd02fa","CBSL":"#fa0202","CBL":"#fa0203","WF2":"#630af2","WF1":"#630af2","CBLF":"#9c959b","CCL":"#9c959b","BS":"#9c959b","TGAS":"#fa0202","Oilshow":"#078238"}}'
    curve_info = json.loads(meta_info)
    curve_info['Unit'].update(new_unit) if len(new_unit) != 0 else curve_info['Unit']
    curve_info['Color'].update(new_color) if len(new_color) != 0 else curve_info['Color']
    log_type = ['LLD', 'LLS', 'DIL', 'ILD', 'ILM', 'RILD', 'RILM', 'HRI', 'RIDPH', 'RIMPH', 'RSFL', 'HDRS', 'HMRS', 'LL3', 'RIPD', 'RIPM', 'AT10', 'AT20', 'AT30', 'AT60', 'AT90', 'AO10', 'AO20', 'AO30', 'AO60', 'AO90', 'AF10', 'AF20', 'AF30', 'AF60', 'AF90', 'AHO10', 'AHO20', 'AHO30', 'AHO60', 'AHO90', 'M2RX', 'M2R9', 'M2R6', 'M2R3', 'M2R1', 'P16H_UNC', 'P22H_UNC', 'P28H_UNC', 'P34H_UNC', 'P40H_UNC', 'A28H_UNC', 'A34H_UNC', 'A40H_UNC', 'P16L_UNC', 'P22L_UNC', 'P28L_UNC', 'P34L_UNC', 'P40L_UNC', 'P22H_UNC_RT', 'P40H_UNC_RT', 'P40H_UNC_R', 'P22H_UNC_R', 'HLLD', 'HLLS', 'RES_DEEP', 'RES_SHAL', 'Rdeep', 'Rshallow', 'P33H_UNC', 'P33L_UNC', 'A33H_UNC', 'A33L_UNC', 'RACHM', 'RACLM', 'RPCHM', 'RPCLM', 'RPCSHM', 'RPCSLM', 'RACSHM', 'RACSLM', 'RAC1HM', 'RAC1LM', 'RPC1HM', 'RPC1LM', 'RAC2HM', 'RAC2LM', 'RPC2HM', 'RPC2LM', 'RAC3HM', 'RAC3LM', 'RPC3HM', 'RPC3LM', 'RAC4HM', 'RAC4LM', 'RPC4HM', 'RPC4LM', 'MSFL', 'RXO8', 'ATR', 'PSR', 'RT', 'RXOZ', 'RSOZ']
    log_type_update = list(set(log_type + log))

    #Define curves in columns list
    if len(curves) != 0:
        curves_list = curves
    else:
        curves_list = data.select_dtypes(include=['float64']).columns.drop(depth_col)
    
    #Calculate numbers of plotting columns and size of whole figure
    cols = len(curves_list); height= 800; width = cols * 200
    #Make subplots layout
    fig = go.Figure()
    fig = make_subplots(rows=1, cols=cols, shared_yaxes=True, horizontal_spacing=0.01)

    #Check selected curves for plotting
    for i, curve in enumerate(curves_list):
        #Assign color code for single curve
        color = curve_info['Color'][curve] if curve in curve_info['Color'].keys() else '#38a2e8'
        unit = curve_info['Unit'][curve] if curve in curve_info['Unit'].keys() else '_'
        #Add trace to subplots
        add_trace(curv=curve, color=color, xaxis=f'x{i+1}', unit=unit, id_subplt=i+1, log=log_type_update)
        
    #Setup the Axes and Layout parameters
    update_xaxes(); update_yaxes(); update_layout([width, height])
    #Show the main figure
    fig.show(config=dict(displayModeBar=True))


import numpy as np
import pandas as pd
import optuna
import lightgbm as lgb
from .score_cal import Score
from sklearn.model_selection import train_test_split
import os
seed = 42


def Train_LGBMC(features: any = None,
                labels: any = None,
                iterations: int = 50,
                scoring: str = None,
                base_score: float = 0.8,  # desired scored
                validation_size: float = 0.2,
                test_size: float = 0.1,
                validation_set: tuple = None,
                test_set: tuple = None,
                task: str = "binary",
                # {0: "valid_score", 1: "train_valid_drop"}
                objectives: any = "valid_score",
                # {0: "min false negative", 1: "min false positive", 2: "balanced"}
                favor_class: any = 1,
                start: any = 1,  # 1 or True
                show_shap: bool = True,
                saved_dir: str = None,
                refit: bool = True,
                ):
    """Construct a function to train a booster.

    Parameters
    ----------
    features : any, {dataframe, array} of predictors
        features use to build model.
    labels : any, {array, Series} of labels
        labels to be predicted.
    iterations : int,
        number of optimization rounds.
    scoring : str,
        scoring method for optimization.
        Available scoring method is sklearn regression metrics: balanced_accuracy, f1_weighted, precision, recall,
                roc_auc, accuracy.
    validation_size: float,
        size of validation set use to evaluate model.
    test_size: float,
        size of test set.
    validation_set: tuple,
        tuple of (X_valid, y_valid)
    test_set: tuple,
        tuple of (X_test, y_test)
    task : str,
        internal method for xgboost to optimize gradient. Available tasks of lightgbm classification are:
        binary, cross_entropy, multiclass, multiclassova.
    objectives : any, {int or str}
        the way to perform optimization.
        "valid_score" or 0: mean to optimize valid score
        "train_valid_drop" or 1: mean to optimize different between train and valid score.
    base_score: float
            control under fitting when apply objective is train_test_drop / 1
    start : any,
        default=1
        is internal param for starting optimize.
    show_shap : bool,
        default=False, show shap chart after tuning if True other while save charts as pictures.

    Returns
    -------
    estimator : object
        The trained model.
    """
    if validation_set and test_set:
        train_x, train_labels = features, labels
        valid_x, valid_labels = validation_set
        test_x, test_labels = test_set

    elif validation_set and test_size:
        valid_x, valid_labels = validation_set
        train_x, test_x, train_labels, test_labels = train_test_split(features,
                                                                      labels,
                                                                      test_size=test_size,
                                                                      random_state=seed)
    elif test_set and validation_size:
        train_x, valid_x, train_labels, valid_labels = train_test_split(features,
                                                                        labels,
                                                                        test_size=validation_size,
                                                                        random_state=seed)
        test_x, test_labels = test_set

    else:
        train_x, valid_x, train_labels, valid_labels = train_test_split(features,
                                                                        labels,
                                                                        test_size=validation_size,
                                                                        random_state=seed)

        train_x, test_x, train_labels, test_labels = train_test_split(features,
                                                                      labels,
                                                                      test_size=test_size,
                                                                      random_state=seed)
    # set datamatrix
    dtrain = lgb.Dataset(train_x, label=train_labels, free_raw_data=False)
    dvalid = lgb.Dataset(valid_x, label=valid_labels, free_raw_data=False)
    dtest = lgb.Dataset(test_x, label=test_labels, free_raw_data=False)

    def objective(trial):
        params = {
            "objective": task,
            "verbosity": -1,
            "feature_pre_filter": False,
            "is_unbalance": trial.suggest_categorical("is_unbalance", [True]),
            "boosting_type": trial.suggest_categorical("boosting_type", ["gbdt"]),
            # "n_estimators": trial.suggest_int("n_estimators", 10, 1000),
            "learning_rate": trial.suggest_loguniform("learning_rate", 1e-3, 0.5),
            "lambda_l1": trial.suggest_loguniform("lambda_l1", 1e-2, 5.0),
            "lambda_l2": trial.suggest_loguniform("lambda_l2", 1e-2, 5.0),
            "num_leaves": trial.suggest_("num_leaves", [21, 31, 41, 51, 61, 63]),
            "feature_fraction": trial.suggest_loguniform("feature_fraction", 0.4, 1.0),
            "bagging_fraction": trial.suggest_loguniform("bagging_fraction", 0.4, 1.0),
            # "bagging_freq": trial.suggest_int("bagging_freq", 1, 15),
            "min_data_in_leaf": trial.suggest_categorical("min_data_in_leaf", [21, 31, 41, 51, 61]),
            "min_sum_hessian_in_leaf": trial.suggest_loguniform("min_sum_hessian_in_leaf", 1e-3, 100),
            # "scale_pos_weight": trial.suggest_loguniform("scale_pos_weight", 1., 100),
            "tree_learner": trial.suggest_categorical("tree_learner", ["serial", "feature"]),
            "pos_bagging_fraction": trial.suggest_loguniform("pos_bagging_fraction", 2e-1, .8),
        }

        model = lgb.train(
            params,
            dtrain,
            # valid_sets=[dvalid],
            # valid_names=["valid"],
            # num_boost_round=500,
            # early_stopping_rounds=5,
            verbose_eval=False)

        # model.fit(train_x, train_labels)#, valid_sets=dvalid, num_boost_round=100)
        preds = model.predict(valid_x)
        pred_labels = preds > 0.5
        pred_trains = model.predict(train_x)
        pred_trains = pred_trains > 0.5
        valid_score = Score(valid_labels, pred_labels,
                            scoring=scoring, favor_class=favor_class)
        train_score = Score(train_labels, pred_trains,
                            scoring=scoring, favor_class=favor_class)
        diff = train_score-valid_score

        if objectives == 0 or objectives == 'valid_score':
            print(f"train-test differs: {diff}")
            return valid_score
        else:
            if valid_score >= base_score:
                print(valid_score)
                return np.abs(diff)
            else:
                return np.exp(np.abs(diff)) * 10

    guides = ["balanced_accuracy", "recall", "f1_weighted", "precision",
              "roc_auc", "R2", "ExV", "Poisson", "D2T", "D2Pi", "D2A"]

    if objectives == 0 or objectives == 'valid_score':
        direction = "maximize"
    else:
        direction = "minimize"

    if start == 1 or start == True:
        study = optuna.create_study(direction=direction,
                                    study_name="find best model")
        study.optimize(objective, n_trials=iterations)

        print("Number of finished trials: {}".format(len(study.trials)))

        print("Best trial:")
        trial = study.best_trial

        print("  Value: {}".format(trial.value))

        print("  Params: ")
        for key, value in trial.params.items():
            print("    {}: {}".format(key, value))

    best_params = {
        "objective": task,
        "metric": "auc",
        "verbosity": -1,
        # "boosting_type": "gbdt",
        "seed": seed,
        "num_boost_round": 1000,
        "early_stopping_rounds": 5,
    }
    best_params.update(study.best_params)

    if refit:
        x_ = pd.concat([train_x, valid_x])
        label_ = pd.concat([train_labels, valid_labels])
        model = lgb.train(params=best_params,
                          train_set=lgb.Dataset(x_,
                                                label=label_,
                                                free_raw_data=False),
                          valid_sets=[dvalid],
                          valid_names=["valid"],
                          verbose_eval=False)
    else:
        model = lgb.train(params=best_params,
                          train_set=dtrain,
                          valid_sets=[dtrain, dvalid],
                          valid_names=["train", "valid"],
                          verbose_eval=False)

    preds = model.predict(valid_x)
    pred_labels = preds > 0.5
    pred_trains = model.predict(train_x)
    pred_trains = pred_trains > 0.5
    pred_tests = model.predict(test_x)
    pred_tests = pred_tests > 0.5

    valid_score = Score(valid_labels,
                        pred_labels,
                        scoring=scoring,
                        favor_class=favor_class)

    train_score = Score(train_labels,
                        pred_trains,
                        scoring=scoring,
                        favor_class=favor_class)
    test_score = Score(test_labels,
                       pred_tests,
                       scoring=scoring,
                       favor_class=favor_class)

    print(f"Train score: {train_score:.3f}")
    if not refit:
        print(f"Valid score: {valid_score:.3f}")

    print(f"Test score: {test_score:.3f}")

    # shap calculation
    feature_names = features.columns
    if saved_dir is None:
        list_dirs = os.scandir(os.getcwd())
        gate_ = [i for i in list_dirs if i.name=="lightgbm_shap"]
        if len(gate_) == 1:
            saved_dir = os.path.join(os.getcwd(), "lightgbm_shap")
        else:
            os.mkdir(os.path.join(os.getcwd(), "lightgbm_shap"))
            saved_dir = os.path.join(os.getcwd(), "lightgbm_shap")

    if show_shap:
        import sys
        sys.path.append('../Shapley')
        from .ShapleyC import shapley_importances_C
        shapley_importances_C(model=model, X=features,
                            feature_names=feature_names,
                            shap_sample_size=1, show_plot=show_shap, saved_dir=saved_dir)

    else:
        import sys
        sys.path.append('../Shapley')
        from .ShapleyC import shapley_importances_C
        shapley_importances_C(model=model, X=features,
                            feature_names=feature_names,
                            shap_sample_size=1, show_plot=False, saved_dir=saved_dir)
    lgb.plot_importance(model)
    lgb.plot_tree(model)

    return model



#Function to create authentication for Sharepoint 365 of Microsoft
from office365.sharepoint.client_context import ClientContext
from office365.runtime.auth.client_credential import ClientCredential
import os
#---------------------------------------------------------------------------------------------------------
class SharepointData:
    def __init__(self, site_url, client_id, client_secret) -> None:
        self.site_url = site_url
        self.client_id = client_id
        self.client_secret = client_secret
        self.ctx = ClientContext(self.site_url).with_credentials(ClientCredential(self.client_id, self.client_secret))
        
    def get_subfolders(self, root_url):
        #Get Folder from server
        folders = self.ctx.web.get_folder_by_server_relative_url(root_url).folders
        #Get Sub-folders
        self.ctx.load(folders).execute_query()
        return folders
    
    def get_files(self, root_url):
        #Get Folder from server
        files = self.ctx.web.get_folder_by_server_relative_url(root_url).files
        #Get files in the current folder
        self.ctx.load(files).execute_query()
        return files
    
    def get_subfolders_url(self, root_url):
        folder_list=[]
        folders = self.ctx.web.get_folder_by_server_relative_url(root_url).folders
        self.ctx.load(folders).execute_query()
        for folder in folders:
            folder_list.append(folder.properties['ServerRelativeUrl'] + '/')
        return folder_list
    
    def get_files_url(self, root_url):
        files_list=[]
        files = self.ctx.web.get_folder_by_server_relative_url(root_url).files
        self.ctx.load(files).execute_query()
        for file in files:
            files_list.append(file.properties['ServerRelativeUrl'])
        return files_list
    
    def download(self, file_url, local_folder_path):
        file_name = file_url.split('/')[-1]
        # file_url is the relative url of the file in sharepoint
        file_path = local_folder_path + file_name
        with open(file_path, "wb") as local_file:
            file = self.ctx.web.get_file_by_server_relative_url(file_url)
            file.download(local_file)
            self.ctx.execute_query()
        print(f"-----Downloaded: {file_name}-----")
        return file_path
    
    def upload(self, file_path, destn_url, remove:bool=False):
        file_name = file_path.split('/')[-1]
        #upload to sharepoint
        destn_folder = self.ctx.web.get_folder_by_server_relative_url(destn_url)
        with open(file_path, 'rb') as content_file:
            file_content = content_file.read()
            destn_folder.upload_file(file_name, file_content).execute_query()
            print(f'{file_name}: Uploaded!')
        if remove:
            os.remove(file_path); print(f'{file_name}: Uploaded & Removed local file!')
            


import numpy as np
import pandas as pd
import optuna
import catboost as cat
from catboost import Pool
from .score_cal import Score
from sklearn.model_selection import train_test_split
import os

seed=42
def Train_CATC(features:any=None,
               labels:any=None,
               iterations:int=50,
               scoring:str=None,
               validation_size:float=0.1,
               test_size:float=0.1,
               validation_set:tuple=None,
               test_set:tuple=None,
               task:str="Logloss",
               objectives:any="valid_score", #{0: "valid_score", 1: "train_test_drop"}
               base_score:float=None,
               favor_class:any=1, #{0: "min false negative", 1: "min false positive", 2: "balanced"}
               start:any=1, # 1 or True
               show_shap:bool=False,
               saved_dir:any=None,
               refit:bool=False,
               ):
    """Construct a function to train a booster.

    Parameters
    ----------
    features : any, {dataframe, array} of predictors
        features use to build model.
    labels : any, {array, Series} of labels
        labels to be predicted.
    iterations : int,
        default = 50, number of optimization rounds.
    scoring : str,
        scoring method for optimization.
        Available scoring method is sklearn regression metrics: balanced_accuracy, f1_weighted, precision, recall,
                roc_auc, accuracy.
    validation_size: float,
        size of validation set use to evaluate model.
    task : str,
        internal method for CatBoost to optimize gradient. Available tasks of CatBoost classification are:
        Logloss, Accuracy, AUC, BalancedAccuracy, F, F1, Precision, Recall, CrossEntropy.
    objectives : any, {int or str}
        the way to perform optimization.
        "valid_score" or 0: mean to optimize valid score
        "train_test_drop" or 1: mean to optimize different between train and test score.
    base_score: float
            control under fitting when apply objective is train_test_drop / 1
    start : any,
        default=1
        is internal param for starting optimize.
    show_shap : bool,
        default=False, show shap chart after tuning if True other while save charts as pictures.

    Returns
    -------
    estimator : object
        The trained model.
    """
    if validation_set and test_set:
        train_x, train_labels = features, labels
        valid_x, valid_labels = validation_set
        test_x, test_labels = test_set
        
    elif validation_set and test_size:
        valid_x, valid_labels = validation_set
        train_x, test_x, train_labels, test_labels = train_test_split(features, 
                                                                      labels, 
                                                                      test_size=test_size, 
                                                                      random_state=seed)
    elif test_set and validation_size:
        train_x, valid_x, train_labels, valid_labels = train_test_split(features, 
                                                                        labels, 
                                                                        test_size=validation_size, 
                                                                        random_state=seed)
        test_x, test_labels = test_set
    else:
        train_x, valid_x, train_labels, valid_labels = train_test_split(features, 
                                                                        labels, 
                                                                        test_size=validation_size, 
                                                                        random_state=seed)
        
        train_x, test_x, train_labels, test_labels = train_test_split(features, 
                                                                      labels, 
                                                                      test_size=test_size, 
                                                                      random_state=seed)
    # if validation_set:
    #     train_x, train_labels = features, labels
    #     valid_x, valid_labels = validation_set
    # else:
    #     train_x, valid_x, train_labels, valid_labels = train_test_split(features, labels, test_size=validation_size, random_state=seed)
    dtrain = Pool(data=train_x, label=train_labels, feature_names=features.columns.to_list())
    dvalid = Pool(data=valid_x, label=valid_labels, feature_names=features.columns.to_list())
    dtest = Pool(data=test_x, label=test_labels, feature_names=features.columns.to_list())

    def objective(trial):
        param = {
            "objective": trial.suggest_categorical("objective", ["Logloss"]),
            #"eval_metric": "Accuracy",
            #"grow_policy": "Lossguide",
            #"bootstrap_type": "Bayesian",
            "bootstrap_type": trial.suggest_categorical("bootstrap_type", ["MVS", "Bernoulli", "Bayesian"]),
            "l2_leaf_reg": trial.suggest_loguniform("l2_leaf_reg", 1e-2, 5),
            "learning_rate": trial.suggest_loguniform("learning_rate", 0.001, 0.3),
            "mvs_reg": trial.suggest_float("mvs_reg", 0, 10),
            "min_data_in_leaf": trial.suggest_int("min_data_in_leaf", 10, 1000),
            "grow_policy": trial.suggest_categorical("grow_policy", ["SymmetricTree", "Lossguide", "Depthwise"]),
            # "nan_mode": trial.suggest_categorical("nan_mode", ["Min", "Max"]),
            # "subsample": trial.suggest_loguniform("subsample", 0.01, 1.),
            # "scale_pos_weight": trial.suggest_loguniform("scale_pos_weight", 1, 20.),
            # "sampling_frequency": trial.suggest_categorical("sampling_frequency", ["PerTree", "PerTreeLevel"]),
            #"auto_class_weights": trial.suggest_categorical("auto_class_weights", ["Balanced", "SqrtBalanced"]),
            #"langevin": trial.suggest_categorical("langevin", [True, False]),
            #"max_ctr_complexity": trial.suggest_int("max_ctr_complexity", 0, 8),
            # "colsample_bylevel": trial.suggest_loguniform("colsample_bylevel", 0.01, 1),
            #"depth": trial.suggest_int("depth", 3, 16)
        }

        if param["objective"] == "Logloss":
            param["auto_class_weights"] = trial.suggest_categorical("auto_class_weights", ["Balanced"])
        # if param["objective"] == "CrossEntropy":
        #     param["scale_pos_weight"] = trial.suggest_loguniform("scale_pos_weight", 1, 20.)
        if param["bootstrap_type"] != "Bayesian":
            param["subsample"] = trial.suggest_loguniform("subsample", 0.4, .9)
        if param["grow_policy"] == "Lossguide":
            param["max_leaves"] = trial.suggest_int("max_leaves", 11, 63)
        if param["grow_policy"] == "SymmetricTree":
            param["boosting_type"] = trial.suggest_categorical("boosting_type", ["Ordered", "Plain"])

        model = cat.train(
            pool=dtrain, 
            params=param, 
            iterations=150,
            # eval_set=dvalid,
            # early_stopping_rounds=5,
            verbose=0,
            )

        preds = model.predict(dvalid, prediction_type="Probability")[:,1]
        pred_labels = preds>0.5
        pred_trains = model.predict(dtrain, prediction_type="Probability")[:,1]
        pred_trains = pred_trains>0.5
        valid_score = Score(valid_labels.astype(int), pred_labels.astype(int), scoring=scoring, favor_class=favor_class)
        train_score = Score(train_labels.astype(int), pred_trains.astype(int), scoring=scoring, favor_class=favor_class)
        diff = train_score-valid_score

        if objectives==0 or objectives=='valid_score':
            print(f"train-test differs: {diff}")
            return valid_score
        else:
            if valid_score >= base_score:
                print(valid_score)
                return np.abs(diff)
            else:
                return np.exp(np.abs(diff)) * 10

    guides = ["balanced_accuracy", "recall", "f1_weighted", "precision", "roc_auc", "R2", "ExV", "Poisson", "D2T", "D2Pi", "D2A"]

    if objectives==0 or objectives=='valid_score':
        direction="maximize"
    else:
        direction="minimize"

    if start==1 or start==True:
        study = optuna.create_study(direction=direction, study_name="find best model")
        study.optimize(objective, n_trials=iterations)

        print("Number of finished trials: {}".format(len(study.trials)))

        print("Best trial:")
        trial = study.best_trial

        print("  Value: {}".format(trial.value))

        print("  Params: ")
        for key, value in trial.params.items():
            print("    {}: {}".format(key, value))

    best_params = {
            #"objective": "Logloss",
            "eval_metric": "Logloss",
        }

    best_params.update(study.best_params)
    
    if refit:
        x_ = pd.concat([train_x, valid_x])
        label_ = pd.concat([train_labels, valid_labels])
        model = cat.train(
            pool=Pool(x_, label=label_), 
            params=best_params, 
            eval_set=dvalid, 
            early_stopping_rounds=5, 
            verbose=0)
    else:
        model = cat.train(
            pool=dtrain, 
            params=best_params, 
            eval_set=dvalid, 
            early_stopping_rounds=5, 
            verbose=0)
        
    preds = model.predict(dvalid, prediction_type="Probability")[:,1]
    pred_labels = preds >0.5
    
    pred_trains = model.predict(dtrain, prediction_type="Probability")[:,1]
    pred_trains = pred_trains >0.5
    
    pred_tests = model.predict(dtest, prediction_type="Probability")[:,1]
    pred_tests = pred_tests >0.5
    
    valid_score = Score(valid_labels.astype(int), 
                        pred_labels.astype(int), 
                        scoring=scoring, 
                        favor_class=favor_class)
    train_score = Score(train_labels.astype(int), 
                        pred_trains.astype(int), 
                        scoring=scoring, 
                        favor_class=favor_class)
    
    test_score = Score(test_labels.astype(int), 
                        pred_tests.astype(int), 
                        scoring=scoring, 
                        favor_class=favor_class)
    
    print(f"Train score: {train_score}")
    if not refit:
        print(f"Valid score: {valid_score}")
        
    print(f"Test score: {test_score}")
    # train_pool_slice = Pool(features, label=labels).slice([i for i in range(len(features.columns))])
    # prediction_diff = model.get_feature_importance(train_pool_slice,
    #                                                type='PredictionDiff',
    #                                                prettified=True)
    #
    # model.plot_predictions(data=train_pool_slice,
    #                        features_to_change=prediction_diff["Feature Id"][:2],
    #                        plot=True,
    #                        plot_file="plot_predictions_file.html")
    if saved_dir is None:
        list_dirs = os.scandir(os.getcwd())
        gate_ = [i for i in list_dirs if i.name=="catboost_shap"]
        if len(gate_) == 1:
            saved_dir = os.path.join(os.getcwd(), "catboost_shap")
        else:
            os.mkdir(os.path.join(os.getcwd(), "catboost_shap"))
            saved_dir = os.path.join(os.getcwd(), "catboost_shap")

    if show_shap:
        import sys
        sys.path.append('../Shapley')
        from .ShapleyC import shapley_importances_C
        shapley_importances_C(model=model, X=features, feature_names=features.columns,
                            shap_sample_size=1, show_plot=show_shap, saved_dir=saved_dir)
    else:
        import sys
        sys.path.append('../Shapley')
        from .ShapleyC import shapley_importances_C
        shapley_importances_C(model=model, X=features, feature_names=features.columns, shap_sample_size=1,
                            show_plot=False, saved_dir=saved_dir)
    return model



import numpy as np
import pandas as pd
import optuna
import lightgbm as lgb
from .score_cal import RScore
from sklearn.model_selection import train_test_split
import os
seed = 42


def Train_LGBMR(features: any = None,
                target: any = None,
                iterations: int = 50,
                scoring: str = None,
                validation_size: float = 0.2,
                test_size: float = 0.1,
                validation_set: tuple = None,
                test_set: tuple = None,
                task: str = "regression",
                # {0: "valid_score", 1: "train_valid_drop"}
                objectives: any = "valid_score",
                base_score:float=None,
                start: any = 1,  # 1 or True
                show_shap: bool = True,
                saved_dir:any=None,
                refit: bool = True,
                ):
    """Construct a function to train a booster.

        Parameters
        ----------
        features : any, {dataframe, array} of predictors
            features use to build model.
        target : any, {array, Series} of target
            continuous target to be predicted.
        iterations : int,
            number of optimization rounds.
        scoring : str,
            scoring method for optimization.
            Available scoring method is sklearn regression metrics: R2 (r2), MAE, MSE, RMSE, MAPE, Poisson, Tweedie, MeAE,
                ExVS, MSLE, ME, Gama, D2T, D2Pi, D2A.
        validation_size: float,
            size of validation set use to evaluate model.
        test_size: float,
            size of test set.
        validation_set: tuple,
            tuple of (X_valid, y_valid)
        test_set: tuple,
            tuple of (X_test, y_test)
        task : str,
            internal method for LightGBM to optimize gradient. Available tasks of LightGBM regression are:
            regression (rmse), mae, huber, fair, poisson, quantile, mape, gamma, tweedie, lambdarank, rank_xendcg,
                rank_xendcg.
        objectives : any, {int or str}
            the way to perform optimization.
            "valid_score" or 0: mean to optimize valid score
            "train_test_drop" or 1: mean to optimize different between train and valid score.
        base_score: float
            control under fitting when apply objective is train_test_drop / 1
        start : any,
            default=1
            is internal param for starting optimize.
        show_shap : bool,
            default=False, show shap chart after tuning if True other while save charts as pictures.

        Returns
        -------
        estimator : object
            The trained model.
    """
    if validation_set and test_set:
        train_x, train_target = features, target
        valid_x, valid_target = validation_set
        test_x, test_target = test_set

    elif validation_set and test_size:
        valid_x, valid_target = validation_set
        train_x, test_x, train_target, test_target = train_test_split(features,
                                                                      target,
                                                                      test_size=test_size,
                                                                      random_state=seed)

    elif test_set and validation_size:
        train_x, valid_x, train_target, valid_target = train_test_split(features,
                                                                        target,
                                                                        test_size=validation_size,
                                                                        random_state=seed)
        test_x, test_target = test_set

    else:
        train_x, valid_x, train_target, valid_target = train_test_split(features,
                                                                        target,
                                                                        test_size=validation_size,
                                                                        random_state=seed)

        train_x, test_x, train_target, test_target = train_test_split(train_x,
                                                                      train_target,
                                                                      test_size=test_size,
                                                                      random_state=seed)

    dtrain = lgb.Dataset(train_x, label=train_target, free_raw_data=False)
    dvalid = lgb.Dataset(valid_x, label=valid_target, free_raw_data=False)

    def objective(trial):
        params = {
            "objective": task,
            "verbosity": -1,
            "feature_pre_filter": False,
            "learning_rate": trial.suggest_loguniform("learning_rate", 1e-2, 0.1),
            "lambda_l1": trial.suggest_loguniform("lambda_l1", 1e-8, 10.0),
            "lambda_l2": trial.suggest_loguniform("lambda_l2", 1e-8, 10.0),
            "num_leaves": trial.suggest_("num_leaves", [21, 31, 41, 51, 61, 63]),
            "feature_fraction": trial.suggest_loguniform("feature_fraction", 0.4, 1.0),
            "bagging_fraction": trial.suggest_loguniform("bagging_fraction", 0.4, 1.0),
            "bagging_freq": trial.suggest_int("bagging_freq", 1, 15),
            "min_data_in_leaf": trial.suggest_categorical("min_data_in_leaf", [21, 31, 41, 51, 61]),
            "tree_learner": trial.suggest_categorical("tree_learner", ["serial", "feature", "data", "voting"])
        }
        if params["objective"] == "tweedie":
            params["tweedie_variance_power"] = trial.suggest_float(
                "tweedie_variance_power", 1., 1.99)

        model = lgb.train(
            params,
            dtrain,
            # valid_sets=[dvalid],
            # valid_names=["valid"],
            # num_boost_round=500,
            # early_stopping_rounds=5,
            verbose_eval=False,
        )

        preds = model.predict(valid_x)
        pred_trains = model.predict(train_x)
        valid_score = RScore(valid_target, preds, scoring=scoring)
        train_score = RScore(train_target, pred_trains, scoring=scoring)
        diff = train_score-valid_score
        if objectives == 0 or objectives == 'valid_score':
            print(f"train-test differs: {diff}")
            return valid_score
        else:
            if valid_score >= base_score:
                print(valid_score)
                return np.abs(diff)
            else:
                return np.exp(np.abs(diff)) * 10

    guides = ["R2", "ExV", "D2T", "D2Pi", "D2A"]

    if scoring in guides:
        if objectives == 0 or objectives == 'valid_score':
            direction = "maximize"
        else:
            direction = "minimize"
    else:
        direction = "minimize"

    if start == 1 or start == True:
        study = optuna.create_study(
            direction=direction, study_name="find best model")
        study.optimize(objective, n_trials=iterations)

        print("Number of finished trials: {}".format(len(study.trials)))

        print("Best trial:")
        trial = study.best_trial

        print("  Value: {}".format(trial.value))

        print("  Params: ")
        for key, value in trial.params.items():
            print("    {}: {}".format(key, value))

    best_params = {
        "objective": task,
        "metric": task,
        "verbosity": -1,
        "boosting_type": "gbdt",
        "seed": 42,
        "num_boost_round": 500,
        "early_stopping_rounds": 5,
    }
    best_params.update(study.best_params)
    if refit:
        x_ = pd.concat([train_x, valid_x])
        target_ = pd.concat([train_target, valid_target])
        model = lgb.train(
            params=best_params,
            train_set=lgb.Dataset(x_, label=target_, free_raw_data=False),
            valid_sets=[dtrain, dvalid],
            valid_names=["train", "valid"],
            verbose_eval=False
        )
    else:
        model = lgb.train(
            params=best_params,
            train_set=dtrain,
            valid_sets=[dtrain, dvalid],
            valid_names=["train", "valid"],
            verbose_eval=False
        )

    preds = model.predict(valid_x)
    pred_trains = model.predict(train_x)
    pred_tests = model.predict(test_x)

    valid_score = RScore(valid_target,
                         preds,
                         scoring=scoring)

    train_score = RScore(train_target,
                         pred_trains,
                         scoring=scoring)

    test_score = RScore(test_target,
                        pred_tests,
                        scoring=scoring)

    print(f"Train score: {train_score}")
    if not refit:
        print(f"Valid score: {valid_score}")
    print(f"Test score: {test_score}")

    feature_names = features.columns
    if saved_dir is None:
        list_dirs = os.scandir(os.getcwd())
        gate_ = [i for i in list_dirs if i.name=="lightgbm_shaps"]
        if len(gate_) == 1:
            saved_dir = os.path.join(os.getcwd(), "lightgbm_shaps")
        else:
            os.mkdir(os.path.join(os.getcwd(), "lightgbm_shaps"))
            saved_dir = os.path.join(os.getcwd(), "lightgbm_shaps")

    if show_shap:
        import sys
        sys.path.append('../Shapley')
        from .ShapleyC import shapley_importances_C
        shapley_importances_C(model=model, X=features,
                            feature_names=feature_names,
                            shap_sample_size=1, show_plot=show_shap, saved_dir=saved_dir)

    else:
        import sys
        sys.path.append('../Shapley')
        from .ShapleyC import shapley_importances_C
        shapley_importances_C(model=model, X=features,
                            feature_names=feature_names,
                            shap_sample_size=1, show_plot=False, saved_dir=saved_dir)

    return model



import shap
from sklearn.model_selection import train_test_split
import matplotlib.pyplot as plt
import os

# create dir to store images
cwd = os.getcwd()
if 'imgs' not in os.listdir():
    path = cwd + '/imgs'
    os.makedirs(path)
else:
    path = cwd + '/imgs'

# plot feature importance by shap
def shapley_importances_R(model:any=None, X:any=None, feature_names:list=None,
                        shap_sample_size:float=0.5, show_plot:bool=False):
    """
        This function use results to visualize features shap values importance for each model
        Inputs must be declared:
            - model: a model to compute shap
            - X: input array / dataframe
            - feature_names: default is None. When you fitted model with dataframe, you don't need to supply feature_names
            feature_names is required when you fitted model by numpy array.
            - shap_sample_size: the size of sample to compute shape valuese
            - show_plot: True or False, default is False (run script). If you run this function in Jupyter
            environment, use 'show_plot=True'
        Outputs:
            - Shap value and dependencies plot.

        Example code:
        >> from features_algos_evaluation import scikit_shapley_importances
        >> shapley_importances(model, X, shap_sample_size=0.1, show_plot=False)
        """

    # Check input
    if X.shape[0] < 10000 or shap_sample_size==1:
        x_shap = X.copy()
    else:
        _, x_shap = train_test_split(X, test_size=shap_sample_size, random_state=42)

    # get model parameters
    model_name = type(model).__name__

    if model_name=='Pipeline':
        model_name = type(model[-1]).__name__
        preprocess = model[:-1]
        x_shap = preprocess.transform(x_shap)
        if model_name =="VotingRegressor":
            model = model.estimator[-1]
        elif model_name=="StackingRegressor":
            model = model.final_estimator
        else:
            model = model[-1]
    else:
        model_name = model_name

    if feature_names is None:
        try:
            feature_names = model.feature_names_in_
        except:
            try:
                feature_names = model.feature_name()
            except:
                feature_names = model.feature_names

    # Calculate Shap value for model
    explainer = shap.Explainer(model, x_shap, feature_names=feature_names)
    shap_values = explainer(x_shap)
    plt.clf()
    shap.summary_plot(shap_values, plot_type='dot', plot_size=(12, 6),
                      feature_names=feature_names, show=show_plot,
                      title=f"{model_name}_summary features importance plot")
    if show_plot is False:
        f = plt.gcf()
        plt.title(f"Feature importances of {model_name} model")
        plt.savefig(f"imgs/{model_name}_summary_plot.png", bbox_inches='tight')
        plt.close()
    for feature in feature_names:
        shap.plots.scatter(shap_values[:, feature], color=shap_values, show=show_plot)
        if show_plot is False:
            f = plt.gcf()
            plt.title(f"{model_name}_{feature}_features dependency plot")
            plt.savefig(f"imgs/{model_name}_{feature}_features dependency plot.png", bbox_inches='tight')
            plt.close()


from ._catboostC import Train_CATC
from ._catboostR import Train_CATR

from ._lightgbmC import Train_LGBMC
from ._lightgbmR import Train_LGBMR

from ._xgboostC import Train_XGBC
from ._xgboostR import Train_XGBR

from .ShapleyC import shapley_importances_C
from .ShapleyR import shapley_importances_R

from .func_score import My_Score
from .score_cal import Score, RScore


__all__ = ['Train_CATC',
            'Train_CATR',
            
            'Train_LGBMC',
            'Train_LGBMR',
            
            'Train_XGBC',
            'Train_XGBR',
            
            'shapley_importances_C',
            'shapley_importances_R',
            
            'My_Score',
            'Score',
            'RScore',
            ]




# DEV FUNCTIONS
from vpitools.Devtools.ShapleyC import shapley_importances_C
from vpitools.Devtools.ShapleyR import shapley_importances_R
from vpitools.Devtools._catboostC import *
from vpitools.Devtools._catboostR import *
from vpitools.Devtools._lightgbmC import *
from vpitools.Devtools._lightgbmR import *
from vpitools.Devtools._xgboostC import *
from vpitools.Devtools._xgboostR import *
from vpitools.Devtools.func_score import *
from vpitools.Devtools.score_cal import *

# EDA FUNCTIONS
from vpitools.eda.plotting.missing_stats import missing_stats
from vpitools.eda.plotting.log_plot import linehist, cross_plot, pair_plot, heatmap, scatter_3d
from vpitools.eda.plotting.view_curves import view_curves


__all__ = [
    'missing_stats',
    'linehist',
    'cross_plot',
    'pair_plot',
    'heatmap',
    'scatter_3d',
    'view_curves',
    'shapley_importances_C',
    'shapley_importances_R',
    ]


from setuptools import setup, find_packages
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()
setup(
    name='vpitools',
    version='1.0.1',
    
    description='VPI Tools',
    url='https://vpi.pvn.vn/',
    
    author='DnA Teams',
    author_email='dna.teams@vpi.pvn.vn',
    license='MIT',
    
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(), include_package_data=True,
    zip_safe=False,
    
    install_requires=['seaborn', 
                    'plotly', 
                    'altair', 
                    'numpy==1.21.6',
                    'matplotlib==3.4.3',
                    'pandas >=1.4.3',
                    'scikit-learn >=1.1.1',
                    'xgboost >= 1.5.1',
                    'lightgbm >= 3.3.2',
                    'matplotlib >= 3.4.3',
                    'shap >= 0.41.0',
                    'fasttreeshap >= 0.1.2',
                    'scipy >= 1.9.0',
                    'mlens >= 0.2.3',
                    'scikit-optimize',
                    'imblearn',
                    'optuna',
                    'catboost',
                    ]
)
